#!/usr/bin/env perl

use strict;
use warnings;
use Getopt::Long;

my $usage = <<'ENDHERE';
NAME:
    uclust_composition.pl
PURPOSE:
    To generate a table of cluster composition by barcode and optionally filter clusters.
INPUT:
    --uc <UC> : UCLUST results (*.uc file)
    --barcodes <Txt|Fasta> : barcode sequences in either Tabular or Fasta format (optional)
    --min <float> : minimum fraction of reads (per barcode) in order to be included in output (default=0.00001 for 0.001%)
OUTPUT:
    --obs <Tsv> : table of cluster and abundance (per barcode if provided)
    --filtered <UC> : filter UC file
AUTHOR/SUPPORT:
    ESKirton@LBL.gov
ENDHERE

my ( $help, $uc_infile, $outfile, $barcodes_infile, $uc_outfile );
my $min = 0.00001;
GetOptions(
	'uc=s'   => \$uc_infile,
    'filtered=s' => \$uc_outfile,
	'outfile=s'  => \$outfile,
	'barcodes=s' => \$barcodes_infile,
	'min=f'      => \$min,
	'help'       => \$help );
if ($help) { print $usage; exit; }
die("Infile required\n") unless $uc_infile and -f $uc_infile;
die("Outfile required\n") unless $outfile;

# LOAD BARCODES
my %barcode_labels   = ();                  # barcode sequence => name
my @barcode_labels   = ('UNKNOWN');         # column order
my %barcode_to_index = ( 'UNKNOWN' => 0 );    # barcode sequence => array index (in @barcode_labels and @obs)
if ($barcodes_infile)
{
	my $barcode;
	my $label;
	my $is_fasta = 0;
	open( IN, "<$barcodes_infile" )
	  or die("Unable to open barcodes file, $barcodes_infile: $!\n");
	while (<IN>)
	{
		chomp;
		next if /^#/;
		next unless $_;
		if (/^>(\S+)/)
		{
			$label    = $1;
			$is_fasta = 1;
			while (<IN>)
			{
				chomp;
				$barcode = $_;
				last;
			}
		}
		elsif ($is_fasta)
		{
			die("Barcodes Fasta file must have entire sequence on only one line.\n");
		}
		else
		{
			( $barcode, $label ) = split(/\t/);
			$label = $barcode unless $label;
		}
		$label =~ s/[_ ]/./g;
		$barcode = uc($barcode);
		push @barcode_labels, $barcode_labels{$barcode} = $label;
		$barcode_to_index{$barcode} = $#barcode_labels;
	}
	close IN;
}

# PARSE UCLUST FILE
# NOTE: BARCODES MAY BE DISCOVERED HERE (I.E. NOT PRESPECIFIED ABOVE)
my $tot_unknown_barcodes = 0;
# $obs[$cluster] = [ obs for barcode 0, obs for barcode 1, ... ] with order matching @labels
my @obs                  = (); 
my ( $index, $cluster, $counter, $barcode_label );
open( IN, "<$uc_infile" ) or die($!);
while (<IN>)
{
	chomp;
	my @row  = split(/\t/);
	my $code = $row[0];
	next unless $code eq 'C' or $code eq 'H';
	$cluster = $row[1];
	$obs[$cluster] = [] if !defined( $obs[$cluster] );
	my $read_id = $row[8];
	if ( $read_id =~ /^\S+#([ATCGN]+)\/?\d?/ )
	{
		my $barcode = uc($1);
		if ( exists( $barcode_labels{$barcode} ) )
		{
			$barcode_label = $barcode_labels{$barcode};
			$index         = $barcode_to_index{$barcode};
		}
		elsif ( ! $barcodes_infile ) 
		{
			push @barcode_labels, $barcode_label = $barcode_labels{$barcode} = $barcode;
			$barcode_to_index{$barcode} = $index = $#barcode_labels;
		}
        else 
        {
            $barcode_label = 'UNKNOWN';
            $index         = 0;
            ++$tot_unknown_barcodes;
        }
    }
	else
	{
		$barcode_label = 'UNKNOWN';
		$index         = 0;
		++$tot_unknown_barcodes;
	}

	# INCREMENT COUNTER
	if ( !defined( $obs[$cluster]->[$index] ) )
	{
		$obs[$cluster]->[$index] = 1;
	}
	else
	{
		$obs[$cluster]->[$index] += 1;
	}
}
close IN;

# DISCARD UNKNOWN CATEGORY IF N/A
unless ($tot_unknown_barcodes)
{
	shift @barcode_labels;
	for ( my $cluster = 0; $cluster <= $#obs; ++$cluster )
	{
		my $a_cluster_obs = $obs[$cluster];
		shift @$a_cluster_obs;
	}
}

# SUM
my @totals = split( //, '0' x scalar(@barcode_labels) );    # total reads per barcode
for ( my $cluster = 0; $cluster <= $#obs; ++$cluster )
{
	for ( my $i = 0; $i <= $#barcode_labels; $i++ )
	{
		$totals[$i] += $obs[$cluster]->[$i] if defined($obs[$cluster]->[$i]);
	}
}

# FILTER BY PERCENT
my @filtered_clusters = ();
my @filtered_obs = ();
for ( my $cluster = 0; $cluster <= $#obs; ++$cluster )
{
	my $okay = 0;
	for ( my $i = 0; $i <= $#barcode_labels; $i++ )
	{
		my $tot = $totals[$i];
		my $ob  = $obs[$cluster]->[$i];
		if ( $ob and $ob / $tot > $min )
		{
			$okay = 1;
			last;
		}
	}
    if ($okay) {
        push @filtered_clusters, $cluster;
        my $cluster_obsAR=$obs[$cluster];
        for ( my $i = 0; $i <= $#barcode_labels; $i++ )
        {
            $cluster_obsAR->[$i] = 0 if !defined( $cluster_obsAR->[$i] );
        }
        push @filtered_obs, $cluster_obsAR;
    }
}

## GENERATE OBSERVED COUNTS TABLE FILE
open( OUT, ">$outfile" ) or die("Unable to write outfile, $outfile: $!\n");
print OUT join( "\t", "#CLUSTER", @barcode_labels ), "\n";
for (my $i=0; $i<=$#filtered_clusters; $i++) {
	print OUT join( "\t", $filtered_clusters[$i], @{$filtered_obs[$i]} ), "\n";
}
close OUT;

## GENERATE UC OUTFILE
my $next_cluster=shift @filtered_clusters;
exit unless defined($next_cluster);
open( IN, "<$uc_infile" ) or die($!);
open(OUT, ">$uc_outfile") or die($!);
while (<IN>)
{
	my @row  = split(/\t/);
	my $code = $row[0];
	next unless $code eq 'C';
	my $cluster = $row[1];
    next unless $cluster eq $next_cluster;
    print OUT $_;
    last unless @filtered_clusters;
    $next_cluster=shift @filtered_clusters;
    last unless defined($next_cluster);
}
close IN;
close OUT;

exit;
